%module ggi

%include system.i


#define        GGI_AUTO     0

#define	ATTR_FGCOLOR	0x0000FF00	/* fgcolor clut index	*/
#define	ATTR_BGCOLOR	0x000000FF	/* bgcolor clut index	*/
#define	ATTR_NORMAL	0x00000000	/* normal style		*/
#define	ATTR_HALF	0x00010000	/* half intensity	*/
#define	ATTR_BRIGHT	0x00020000	/* high intensity	*/
#define	ATTR_INTENSITY	0x00030000	/* mask to get intensity*/
#define	ATTR_UNDERLINE	0x00040000	/* underline attribute	*/
#define	ATTR_BOLD	0x00080000	/* bold style		*/
#define	ATTR_ITALIC	0x00100000	/* italic style		*/
#define	ATTR_REVERSE	0x00200000	/* reverse fg/bg	*/
#define	ATTR_BLINK	0x00800000	/* enable blinking	*/
#define	ATTR_FONT	0xFF000000	/* font table		*/


#define	GGI_COLOR_PRECISION 16	/* 16 bit per R,G, B value	*/


#define GT_DEPTH_SHIFT		(0)
#define GT_SIZE_SHIFT		(8)
#define GT_SUBSCHEME_SHIFT	(16)
#define GT_SCHEME_SHIFT		(24)

#define GT_DEPTH_MASK		(0xff << GT_DEPTH_SHIFT)
#define GT_SIZE_MASK		(0xff << GT_SIZE_SHIFT)
#define GT_SUBSCHEME_MASK	(0xff << GT_SUBSCHEME_SHIFT)
#define GT_SCHEME_MASK		(0xff << GT_SCHEME_SHIFT)


/* Bitmeaning defines */

#define GGI_BM_TYPE_NONE	0x000000	/* This bit is not in use */

/*
  Bit influences color of displayed pixel
*/
#define GGI_BM_TYPE_COLOR	0x010000

#define		GGI_BM_SUB_RED		0x0100
#define		GGI_BM_SUB_GREEN	0x0200
#define		GGI_BM_SUB_BLUE		0x0300

#define		GGI_BM_SUB_CYAN		0x1000
#define		GGI_BM_SUB_MAGENTA	0x1100
#define		GGI_BM_SUB_YELLOW	0x1200
#define		GGI_BM_SUB_K		0x1300

#define		GGI_BM_SUB_Y		0x2000
#define		GGI_BM_SUB_U		0x2100
#define		GGI_BM_SUB_V		0x2200

#define		GGI_BM_SUB_CLUT		0xf000	/* This bit Color or attrib ? */

/*
  Bit changes appearance of pixel/glyph
*/
#define GGI_BM_TYPE_ATTRIB	0x020000

#define 	GGI_BM_SUB_ALPHA		0x0100

#define 	GGI_BM_SUB_BLINK		0x1000
#define		GGI_BM_SUB_INTENSITY		0x1100
#define		GGI_BM_SUB_UNDERLINE		0x1200
#define		GGI_BM_SUB_BOLD			0x1300
#define		GGI_BM_SUB_ITALIC		0x1400

#define		GGI_BM_SUB_FGCOL		0x2000
#define		GGI_BM_SUB_BGCOL		0x2100

#define		GGI_BM_SUB_TEXNUM		0x3000
#define		GGI_BM_SUB_FONTSEL		0x3100  /*select different font banks*/
#define		GGI_BM_SUB_PALSEL		0x3200	/* select different palettes */
#define		GGI_BM_SUB_MODESEL		0x3300	/* select different palettes */

/*
  Bit that influence drawing logic
*/
#define GGI_BM_TYPE_LOGIC	0x030000

#define		GGI_BM_SUB_ZBUFFER		0x0100
#define		GGI_BM_SUB_WRITEPROT		0x1000
#define		GGI_BM_SUB_WINDOWID		0x2000

/* Enumerated schemes */
#define GT_TEXT			((0x01) << GT_SCHEME_SHIFT)
#define GT_TRUECOLOR		((0x02) << GT_SCHEME_SHIFT)
#define GT_GREYSCALE		((0x03) << GT_SCHEME_SHIFT)
#define GT_PALETTE		((0x04) << GT_SCHEME_SHIFT)
#define GT_STATIC_PALETTE	((0x05) << GT_SCHEME_SHIFT)

/* Subschemes */
#define GT_SUB_REVERSE_ENDIAN   ((0x01) << GT_SUBSCHEME_SHIFT)
#define GT_SUB_HIGHBIT_RIGHT    ((0x02) << GT_SUBSCHEME_SHIFT)
#define GT_SUB_PACKED_GETPUT    ((0x04) << GT_SUBSCHEME_SHIFT)



/* Common graphtypes */
#define GT_TEXT16	GT_CONSTRUCT(4, GT_TEXT, 16)
#define GT_TEXT32	GT_CONSTRUCT(8, GT_TEXT, 32)
#define GT_1BIT		GT_CONSTRUCT(1, GT_PALETTE, 1)
#define GT_2BIT		GT_CONSTRUCT(2, GT_PALETTE, 2)
#define GT_4BIT		GT_CONSTRUCT(4, GT_PALETTE, 4)
#define GT_8BIT		GT_CONSTRUCT(8, GT_PALETTE, 8)
#define GT_15BIT	GT_CONSTRUCT(15, GT_TRUECOLOR, 16)
#define GT_16BIT	GT_CONSTRUCT(16, GT_TRUECOLOR, 16)
#define GT_24BIT	GT_CONSTRUCT(24, GT_TRUECOLOR, 24)
#define GT_32BIT	GT_CONSTRUCT(24, GT_TRUECOLOR, 32)
#define GT_AUTO		(0)
#define GT_INVALID	(0xffffffff)



#define GGI_CMDFLAG_LIBGGI (GII_CMDFLAG_EXTERNAL>>1)

/* Tell target that the application should not/should be halted when the
   display is unmapped.	The default is to halt the application.
*/
#define GGICMD_NOHALT_ON_UNMAP (GII_CMDFLAG_EXTERNAL | GGI_CMDFLAG_LIBGGI \
				| GII_CMDFLAG_NODATA | 0x01)
#define GGICMD_HALT_ON_UNMAP (GII_CMDFLAG_EXTERNAL | GGI_CMDFLAG_LIBGGI \
			      | GII_CMDFLAG_NODATA | 0x02)

/* Requests the application to switch target/mode, or to stop drawing on
   the visual.
   The latter is only sent if the application has explicitly requested
   GGICMD_NOHALT_ON_UNMAP. When a GGI_REQSW_UNMAP request is sent the
   application should respond by sending a GGICMD_ACKNOWLEDGE_SWITCH event
   as quickly as possible. After the acknowledge event is sent the
   application must not draw onto the visual until it recieves an evExpose
   event, which tells the application that the visual is mapped back again.
*/

#define GGICMD_REQUEST_SWITCH (GII_CMDFLAG_EXTERNAL \
			       | GGI_CMDFLAG_LIBGGI | 0x01)

/* Used for 'request' field in ggi_cmddata_switchrequest */
#define GGI_REQSW_UNMAP		0x01
#define GGI_REQSW_MODE		0x02
#define GGI_REQSW_TARGET	0x04

#define GGICMD_ACKNOWLEDGE_SWITCH (GII_CMDFLAG_EXTERNAL | GGI_CMDFLAG_LIBGGI \
				   | GII_CMDFLAG_NODATA | 0x03)

%typemap (python,out) PyObject *{
	$target = $source;
	}

%typemap (python,in) PyObject *{
	$target = $source;
	}




typedef enum {

	evNothing = 0,	/* event is not valid. (must be zero)	*/

	evCommand,	/* report command/do action		*/
	evInformation,	/* notification of new information	*/

	evExpose,	/* exposure event			*/	
	/* empty slot */

	evKeyPress=5,	/* key has been pressed			*/
	evKeyRelease,	/* key has been released		*/
	evKeyRepeat,	/* automatically repeated keypress	*/

	evPtrRelative,	/* pointer movements reported relative	*/
	evPtrAbsolute,	/* pointer movements reported absolute	*/
	evPtrButtonPress,	/* pointer button pressed	*/
	evPtrButtonRelease,	/* pointer button released	*/

	evValRelative,	/* valuator change (reported relative)	*/
	evValAbsolute,	/* valuator change (reported absolute)	*/

	evLast		/* must be less than 33			*/

}  gii_event_type;

//#define	ev(x)	em ## x = (1 << ev ## x)
typedef enum {

	emNothing = 1 << evNothing,

	emCommand = 1 << evCommand,
	emInformation = 1 << evInformation,
	emExpose = 1 << evExpose,

	emKeyPress = 1 << evKeyPress,
	emKeyRelease = 1 << evKeyRelease,
	emKeyRepeat = 1 << evKeyRepeat,
	emKey	= emKeyPress | emKeyRelease | emKeyRepeat,
	emKeyboard	= emKey,

	emPtrRelative = 1 << evPtrRelative,
	emPtrAbsolute = 1 << evPtrAbsolute,
	emPtrButtonPress = 1 << evPtrButtonPress,
	emPtrButtonRelease = 1 << evPtrButtonRelease,
	emPtrMove	= emPtrRelative | emPtrAbsolute,
	emPtrButton	= emPtrButtonPress | emPtrButtonRelease,
	emPointer	= emPtrMove | emPtrButton,

	emValRelative = 1 << evValRelative,
	emValAbsolute = 1 << evValAbsolute,
	emValuator	= emValRelative | emValAbsolute,

	emZero  = 0,
	emAll	= ((1 << evLast) - 1) & ~emNothing

} gii_event_mask;


#define GII_EV_ORIGIN_NONE		0x00000000	/* Anonymous */
#define GII_EV_ORIGIN_SENDEVENT		0x80000000	/* Event was due to SendEvent call */
//#define GII_EV_ORIGIN(id)		(id)		/* Otherwise, it's an id */
		
#define GII_EV_TARGET_ALL		0x00000000	/* broadcast event */

#define GII_CMDFLAG_NODATA	(1<<31)	/* Event has no data */
#define GII_CMDFLAG_PRIVATE	(1<<30) /* The code is specific to a certain
					   inputlib */
#define GII_CMDFLAG_EXTERNAL	(1<<29)	/* Event is sent to/from an external
					   system (like LibGGI) */
#define GII_CMD_DATA_MAX  (248-sizeof(gii_cmd_nodata_event))


/* Pixelformat for ggiGet/Put* buffers and pixellinearbuffers */
struct ggi_pixelformat {

	int		depth;		/* Number of significant bits */
	int		size;		/* Physical size in bits */

	/* 
	 * Simple and common things first :
	 * 
	 * Usage of the mask/shift pairs:
	 * If new_value is the _sizeof(ggi_pixel)*8bit_ value of the thing 
	 * you want to set, you do
	 *
	 * *pointer &= ~???_mask; 		// Mask out old bits 
	 * *pointer |= (new_value>>shift) & ???_mask;
	 * 
	 * The reason to use 32 bit and "downshifting" is alignment
	 * and extensibility. You can easily adjust to other datasizes
	 * with a simple addition ...
	 */
	
	/* Simple colors:
	 */
	long	red_mask;	/* Bitmask of red bits */
	int		red_shift;	/* Shift  for red bits */

	long	green_mask;	/* Bitmask of green bits */
	int		green_shift;	/* Shift  for green bits */

	long	blue_mask;	/* Bitmask of blue bits */
	int		blue_shift;	/* Shift  for blue bits */

	/* A few common attributes :
	 */
	long	alpha_mask;	/* Bitmask of alphachannel bits */
	int		alpha_shift;	/* Shift  for alpha bits */

	long	clut_mask;	/* Bitmask of bits for the clut */
	int		clut_shift;	/* Shift  for bits for the clut*/

	long	fg_mask;	/* Bitmask of foreground color */
	int		fg_shift;	/* Shift  for foreground color */

	long	bg_mask;	/* Bitmask of background color */
	int		bg_shift;	/* Shift  for background color */

	long	texture_mask;	/* Bitmask of the texture (for
					   textmodes - the actual character) */
	int		texture_shift;	/* Shift  for texture */

	/*
	 * Now if this doesn't suffice you might want to parse the following
	 * to find out what each bit does:
	 */

	long		bitmeaning[sizeof(ggi_pixel)*8];

	/* Shall we keep those ?
	 */
	long		flags;		/* Pixelformat flags */

	long		stdformat;	/* Standard format identifier */
	/* This one has only one use for the usermode application:
	 * To quickly check, if two buffers are identical. If both
	 * stdformats are the same and _NOT_ 0 (which means "WEIRD"),
	 * you may use things like memcpy between them which will have
	 * the desired effect ...
	 */
	
};


struct ggi_color{ 
	int r,g,b,a;
	};

struct ggi_coord {
	int x, y;
	};

/*
 * ggi_mode structure
 */
typedef struct		/* requested by user and changed by driver    */
{
	long		frames;		/* frames needed		    */
	ggi_coord	visible;	/* vis. pixels, may change slightly */
	ggi_coord	virt;		/* virtual pixels, may change	    */
	ggi_coord	size;		/* size of visible in mm	    */
	long	graphtype;	/* which mode ?			    */
	ggi_coord	dpp;		/* dots per pixel		    */

	ggi_mode() {}

} ggi_mode;



/* Pixelformat flags */
#define GGI_PF_REVERSE_ENDIAN	0x01
#define GGI_PF_HIGHBIT_RIGHT	0x02
#define GGI_PF_HAM		0x04
#define GGI_PF_EXTENDED		0x08



struct timeval
{
	long tv_sec;
	long tv_usec;

	timeval() {}
};

/*	this information is reported with all events. Use the <any> field
**	in a gii_event structure to access these fields.
*/ 
typedef struct {

	uint8	size;		/* size of event in bytes	*/
	uint8	type;		/* type of this event		*/
	sint16  error;		/* error (for replies)		*/
	uint32	origin;		/* origin device (etc)		*/
	uint32	target;		/* target device (etc)		*/
	struct timeval time;	/* timestamp			*/

} gii_any_event;





/*	These are used internally or to/from the application.
	The same event is used for both Command and Information	events.
	The recipient must not store references to the data. If	the data
	information is needed afterwards, copy it!
*/

typedef struct {

//	COMMON_DATA;
	uint8	size;		/* size of event in bytes	*/
	uint8	type;		/* type of this event		*/
	sint16  error;		/* error (for replies)		*/
	uint32	origin;		/* origin device (etc)		*/
	uint32	target;		/* target device (etc)		*/
	struct timeval time;	/* timestamp			*/

	uint32	code;		/* command/request code		*/

} gii_cmd_nodata_event;


typedef struct {

//	COMMON_DATA;
	uint8	size;		/* size of event in bytes	*/
	uint8	type;		/* type of this event		*/
	sint16  error;		/* error (for replies)		*/
	uint32	origin;		/* origin device (etc)		*/
	uint32	target;		/* target device (etc)		*/
	struct timeval time;	/* timestamp			*/

	uint32	code;			/* command/request code */
	uint8	data[GII_CMD_DATA_MAX];	/* command related data */ 

} gii_cmd_event;


/* This one is posted, if at least one event has been lost due to queue 
   overflow: */
#define GII_CMDCODE_EVENTLOST  (0x01 | GII_CMDFLAG_NODATA)

/* This event is sent/received to require/get the capabilities of a device
 * as specified in target/origin.
 * An event stating num_buttons=num_axes=0 says, that the device is inactive,
 * unplugged, whatever. Devices automatically report (detectable) state
 * changes via devinfo. But you need to re-query the valinfo records.
 */
#define GII_CMDCODE_GETDEVINFO	(0x01)
typedef struct {
	
	char		longname[75];
	char		shortname[5];
	
	gii_event_mask	can_generate;

	uint32		num_buttons;	/* Maximum number of buttons. */
	uint32		num_axes;	/* Maximum number of axes. */

} gii_cmddata_getdevinfo;

typedef struct gii_valrange {
	sint32		min, center, max;
} gii_valrange;

typedef enum { 
	GII_PT_UNKNOWN,			/* unknown */
	GII_PT_TIME, 			/* base unit s */
	GII_PT_FREQUENCY,		/* base unit 1/s (Hz) */
	GII_PT_LENGTH, 			/* base unit m */
	GII_PT_VELOCITY, 		/* base unit m/s */
	GII_PT_ACCELERATION, 		/* base unit m/s^2 */
	GII_PT_ANGLE,			/* base unit radian */
	GII_PT_ANGVELOCITY,		/* base unit radian/s */
	GII_PT_ANGACCELERATION,		/* base unit radian/s^2 */
	GII_PT_AREA, 			/* base unit m^2 */
	GII_PT_VOLUME, 			/* base unit m^3 */
	GII_PT_MASS, 			/* base unit kg */
	GII_PT_FORCE, 			/* base unit N (kg*m/s^2) */
	GII_PT_PRESSURE, 		/* base unit N/m^2 (Pa) */
	GII_PT_TORQUE,	 		/* base unit Nm */
	GII_PT_ENERGY,			/* base unit Nm, VAs, J */
	GII_PT_POWER,			/* base unit Nm/s, VA, W */
	GII_PT_TEMPERATURE,		/* base unit K */
	GII_PT_CURRENT,			/* base unit A */
	GII_PT_VOLTAGE,			/* base unit V (kg*m^2/(As^3)) */
	GII_PT_RESISTANCE, 		/* base unit V/A (Ohm) */
	GII_PT_CAPACITY, 		/* base unit As/V (Farad) */
	GII_PT_INDUCTIVITY, 		/* base unit Vs/A (Henry) */
	GGI_PT_LAST
} gii_phystype;

/* This event is sent/received to require/get detailed data about valuator
 * axis.
 */
#define GII_CMDCODE_GETVALINFO	(0x02)
#define GII_VAL_QUERY_ALL 0xffffffff
typedef struct {
	
	uint32		number;		/* Number of the valuator */
/* You may send this to get all valuators at once. Not recommended. */

	char		longname[75];
	char		shortname[5];
	
	gii_valrange	range;
	gii_phystype	phystype;

	/* The SI value can be computed using the values below
	 * and the following formula (unless SI_mul is 0, which
	 * means the device is non-linear or the factor is unknown).
	 *
	 *   float SI;
	 *
	 *   SI = (float) (SI_add + value[n]) * (float) SI_mul
	 *        / (float) SI_div * pow(2.0, SI_shift);
	 */

       	sint32	SI_add,SI_mul,SI_div,SI_shift;

} gii_cmddata_getvalinfo;


/*
******************************************************************************
 Other events
******************************************************************************
*/

/*	Exposure events give rectangles that need to be refreshed.
*/
typedef struct {

//	COMMON_DATA;
	uint8	size;		/* size of event in bytes	*/
	uint8	type;		/* type of this event		*/
	sint16  error;		/* error (for replies)		*/
	uint32	origin;		/* origin device (etc)		*/
	uint32	target;		/* target device (etc)		*/
	struct timeval time;	/* timestamp			*/

	uint32	x,y;
	uint32	h,w;

} gii_expose_event;


/*	key events should be used to report events obtained from keys and
**	other switches. 
*/
typedef struct {

//	COMMON_DATA;
	uint8	size;		/* size of event in bytes	*/
	uint8	type;		/* type of this event		*/
	sint16  error;		/* error (for replies)		*/
	uint32	origin;		/* origin device (etc)		*/
	uint32	target;		/* target device (etc)		*/
	struct timeval time;	/* timestamp			*/

	uint32	modifiers;	/* current modifiers in effect */
	uint32	sym;		/* meaning of key	*/
	uint32  label;		/* label on key		*/
	uint32  button;		/* button number	*/

} gii_key_event;

/*	This is used to report change of pointer position. 
**	Depending on the event type, the values are either absolute
**	or relative.
*/
typedef struct {

//	COMMON_DATA;
	uint8	size;		/* size of event in bytes	*/
	uint8	type;		/* type of this event		*/
	sint16  error;		/* error (for replies)		*/
	uint32	origin;		/* origin device (etc)		*/
	uint32	target;		/* target device (etc)		*/
	struct timeval time;	/* timestamp			*/

	sint32	x, y;		/* absolute/relative position	*/
	sint32  z, wheel;

} gii_pmove_event;

/*	Button events are sent to report a change in pointer button
**	state.  Depending on the event type, the button is either being
**	pressed or released.
*/
typedef struct {

//	COMMON_DATA;
	uint8	size;		/* size of event in bytes	*/
	uint8	type;		/* type of this event		*/
	sint16  error;		/* error (for replies)		*/
	uint32	origin;		/* origin device (etc)		*/
	uint32	target;		/* target device (etc)		*/
	struct timeval time;	/* timestamp			*/

	uint32	button;		/* button number. This is a number, NOT a mask */

} gii_pbutton_event;

#define GII_PBUTTON_(x)		(x)	/* Generic button access */

#define GII_PBUTTON_NONE	0	/* Dummy */

#define GII_PBUTTON_LEFT	1	/* Left or primary button */
#define GII_PBUTTON_PRIMARY	1
#define GII_PBUTTON_FIRST	1

#define GII_PBUTTON_RIGHT	2	/* Right/Secondary button */
#define GII_PBUTTON_SECONDARY	2
#define GII_PBUTTON_SECOND	2

#define GII_PBUTTON_MIDDLE	3	/* Middle or tertiary  */
#define GII_PBUTTON_TERTIARY	3
#define GII_PBUTTON_THIRD	3

/*	Valuator events report a change of up to 32 of the
**	input device valuators. Only a range of 32 valuators beginning
**	from first can be reported with one event.
*/
typedef struct {

//	COMMON_DATA;
	uint8	size;		/* size of event in bytes	*/
	uint8	type;		/* type of this event		*/
	sint16  error;		/* error (for replies)		*/
	uint32	origin;		/* origin device (etc)		*/
	uint32	target;		/* target device (etc)		*/
	struct timeval time;	/* timestamp			*/

	uint32	first;		/* first valuator reported	*/
	uint32	count;		/* number reported		*/
	sint32	value[32];	/* absolute/relative values	*/

} gii_val_event;

typedef union {

	uint8			size;		/* size of this event	*/

	gii_any_event		any;		/* access COMMON_DATA	*/
	gii_cmd_event		cmd;		/* command/information	*/
	gii_expose_event	expose;		/* exposure event	*/
	gii_val_event		val;		/* valuator change	*/
	gii_key_event		key;		/* key press/release	*/
	gii_pmove_event		pmove;		/* pointer move		*/
	gii_pbutton_event	pbutton;	/* pointer buttons	*/


	ggi_event() {}	

} ggi_event;






%inline %{

#include <ggi.h>
#include <errors.h>





//////////////////////////////////////////////////////////////
//
// ------- a trick to convert macros to python functions
//
#undef ATTR_COLOR
long ATTR_COLOR(int fg, int bg) 
{
	return ((bg & 0xFF) | ((fg & 0xFF) << 8));
} 

/* Swap the bytes in a 16 respective 32 bit unsigned number */
#undef GGI_BYTEREV16
long GGI_BYTEREV16(int x) { return ((x)<<8 | (x)>>8); }

#undef GGI_BYTEREV32
long GGI_BYTEREV32(int x)  {
	return ((x)<<24 | ((x)&0xff00)<<8 | ((x)&0xff0000)>>8 | (x)>>24);
	}

/* Swap the bitgroups in an 8 bit unsigned number */
#undef GGI_BITREV4
int GGI_BITREV4(int x) { return (((x)>>4) | ((x)<<4)); }

#undef GGI_BITREV2
int GGI_BITREV2(int x) {
	return ((x)>>6 | ((x)&0x30)>>2 | ((x)&0x0c)<<2 | (x)<<6);
	}

#undef GGI_BITREV1
int GGI_BITREV1(int x)  {
	return ((x)>>7 | ((x)&0x40)>>5 | ((x)&0x20)>>3 | ((x)&0x10)>>1 | ((x)&0x08)<<1 | ((x)&0x04)<<3 | ((x)&0x02)<<5 | (x)<<7);
	}

/* Macros to set info in a ggi_graphtype. */
#undef GT_SETDEPTH
long GT_SETDEPTH(long gt, long x){
    gt = ((gt) & ~GT_DEPTH_MASK) | ((x)<<GT_DEPTH_SHIFT); 
    return gt;	
    }

#undef GT_DEPTH
long GT_DEPTH(long x) {
	return (((x) & GT_DEPTH_MASK) >> GT_DEPTH_SHIFT);
	}

#undef GT_SIZE
long GT_SIZE(long x) {
	return 	(((x) & GT_SIZE_MASK) >> GT_SIZE_SHIFT);
	}

#undef GT_SUBSCHEME
long GT_SUBSCHEME(long x){
	return ((x) & GT_SUBSCHEME_MASK);
	}

#undef GT_SCHEME
long GT_SCHEME(long x) {
	return 	((x) & GT_SCHEME_MASK);
	}


#undef GT_SETSIZE
long GT_SETSIZE(long gt, long x) {
   gt = ((gt) & ~GT_SIZE_MASK) | ((x)<<GT_SIZE_SHIFT);
   return gt;
   }

#undef GT_SETSUBSCHEME
long GT_SETSUBSCHEME(long gt, long x) {
   gt = ((gt) & ~GT_SUBSCHEME_MASK) | (x); 
   return gt;
   }

#undef GT_SETSCHEME
long GT_SETSCHEME(long gt, long x) {
   gt = ((gt) & ~GT_SCHEME_MASK) | (x);
   return gt;
   }

/* Macro that constructs a graphtype */
#undef GT_CONSTRUCT
long GT_CONSTRUCT(long depth, long scheme, long size) {
  return ((depth) | (scheme) | ((size) << GT_SIZE_SHIFT));
  }


ggi_color create_color(int r=0,int g=0,int b=0,int a=0) {
	ggi_color c;
	c.r=r; c.g=g; c.b=b; c.a=a;
	return c;
	}

ggi_coord create_coord(int x=0,int y=0) {
	ggi_coord c;
	c.x=x; c.y=y;
	return c;
	}


void E(char *msg)
{
	PyErr_SetString(PyExc_Exception, msg);
}

// check error codes
int check(int err) 
{
	switch(err)
	{
	case  GGI_OK:
		break;	/* All is well */
	case  GGI_ENOMEM:
		E("Out of memory");
		break;
	case  GGI_ENOFILE:
		E("File does not exist");
		break;
	case  GGI_ENODEVICE:
		E("Input/Output device can not be opened");
		break;
	case  GGI_EARGREQ:
		E("Required argument missing");
		break;
	case  GGI_EARGINVAL:
		E("Invalid argument(s)");
		break;
	case  GGI_ENOTALLOC:
		E("Trying to use or deallocate a resource that was not previously allocated");
		break;
	case  GGI_EFATAL:
		E("Fatal error - the state of the target of the operation is undefined");
		break;
	case  GGI_ELOCKBUSY:
		E("Lock is in use");
		break;
	case  GGI_ENOTFOUND:
		E("The requested object was not found");
		break;
	case  GGI_EEXCLUSIVE:
		E("Tried to get non-exclusive access to object which only supports exclusive access");
		break;
	case  GGI_EEVUNKNOWN:
		E("Unknown event sent to input source");
		break;
	case  GGI_EEVNOTARGET:
		E("No apropriate target for sent event");
		break;
	case  GGI_EEVOVERFLOW:
		E("Overflow when queuing event");
		break;
	case  GGI_EUNKNOWN:
		E("Unknown error");
		break;
	}
	return err;
}

/*
class color
{
public:
	ggi_color self;

	color() {}

	long getr() { return self.r; }
	long getg() { return self.g; }
	long getb() { return self.b; }
	long geta() { return self.a; }

	void setr(long p) { self.r = p; }
	void setg(long p) { self.g = p; }
	void setb(long p) { self.b = p; }
	void seta(long p) { self.a = p; }

};
*/

/*
class coord
{
public:
	ggi_coord self;
	
	
	coord(int x, int y) {
		self.x = x;
		self.y = y;
		}

	int getx() { return self.x; }
	int gety() { return self.y; }

	int setx(int x) { self.x = x; }
	int sety(int y) { self.y = y; }
};
*/



/*
class mode
{
public:
	ggi_mode tm;
	ggi_visual_t vis;

	
	// access routines
	long getframes() {
		return tm.frames; 
		}
	void setframes(long p) {
		tm.frames = p; 
		}
	
	ggi_coord getvisible() {
		return tm.visible;
		}
	void setvisible(ggi_coord *c){
		tm.visible.x = c->x;
		tm.visible.y = c->y;	
		}

	ggi_coord getvirt() {
		return tm.virt;
		}
	void setvirt(ggi_coord *c){
		tm.virt.x = c->x;
		tm.virt.y = c->y;	
		}

	ggi_coord getsize() {
		return tm.size;
		}
	void setsize(ggi_coord *c){
		tm.size.x = c->x;
		tm.size.y = c->y;	
		}


	long getgraphtype() {
		return tm.graphtype;
		}
	void setgraphtype(long p){
		tm.graphtype = p;
		}

	ggi_coord getdpp() {
		return tm.dpp;
		}
	void setdpp(ggi_coord *c){
		tm.dpp.x = c->x;
		tm.dpp.y = c->y;	
		}



};
*/


/*
class event
{
public:
	ggi_event ev;

	event() {}

	PyObject *getany() {
		PyObject *list = PyList_New(0);

		PyList_Append(list, PyInt_FromLong(ev.any.size));
		PyList_Append(list, PyInt_FromLong(ev.any.type));
		PyList_Append(list, PyInt_FromLong(ev.any.error));
		PyList_Append(list, PyInt_FromLong(ev.any.origin));
		PyList_Append(list, PyInt_FromLong(ev.any.target));
		PyList_Append(list, PyInt_FromLong(ev.any.time.tv_sec));
		PyList_Append(list, PyInt_FromLong(ev.any.time.tv_usec));
		return list;
		}
	void setany(PyObject *list) {
		ev.any.size = PyInt_AsLong(PyList_GetItem(list,0));
		ev.any.type = PyInt_AsLong(PyList_GetItem(list,1));
		ev.any.error = PyInt_AsLong(PyList_GetItem(list,2));
		ev.any.origin = PyInt_AsLong(PyList_GetItem(list,3));
		ev.any.target = PyInt_AsLong(PyList_GetItem(list,4));
		ev.any.time.tv_sec = PyInt_AsLong(PyList_GetItem(list,5));
		ev.any.time.tv_usec = PyInt_AsLong(PyList_GetItem(list,6));
		}

	PyObject *getcmd() {
		PyObject *list = getany();
		PyList_Append(list, PyInt_FromLong(ev.cmd.code));
		PyList_Append(list, PyString_FromStringAndSize(
			(char *) ev.cmd.data, GII_CMD_DATA_MAX));
		return list;
		}
	void setcmd(PyObject *list) {
		setany(list);
		ev.cmd.code = PyInt_AsLong(PyList_GetItem(list,7));
		strncpy( (char *) ev.cmd.data,
			(char*) PyString_AsString(
				PyList_GetItem(list,8) ),
			GII_CMD_DATA_MAX);		
		}

	PyObject *getval() {
		PyObject *list = getany();
		PyList_Append(list, PyInt_FromLong(ev.val.first));
		PyList_Append(list, PyInt_FromLong(ev.val.count));
		PyObject *value = PyList_New(32);
		for(int i = 0; i < 32; i++)
			PyList_SetItem(value,i,
				PyInt_FromLong(ev.val.value[i]) );
		PyList_Append(list, value);
		return list;
		}
	void setval(PyObject *list){
		setany(list);
		ev.val.first = PyInt_AsLong(PyList_GetItem(list,7));
		ev.val.count = PyInt_AsLong(PyList_GetItem(list,8));
		PyObject *value = PyList_GetItem(list,9);
		for(int i = 0; i < 32; i++)
			ev.val.value[i] = PyInt_AsLong( 
				PyList_GetItem(value,i) );
		}

	PyObject *getexpose() {
		PyObject *list = getany();
		PyList_Append(list, PyInt_FromLong(ev.expose.x));
		PyList_Append(list, PyInt_FromLong(ev.expose.y));
		PyList_Append(list, PyInt_FromLong(ev.expose.h));
		PyList_Append(list, PyInt_FromLong(ev.expose.w));			
		return list;
		}
	void setexpose(PyObject *list){
		setany(list);
		ev.expose.x = PyInt_AsLong(PyList_GetItem(list,7));
		ev.expose.y = PyInt_AsLong(PyList_GetItem(list,8));
		ev.expose.h = PyInt_AsLong(PyList_GetItem(list,9));
		ev.expose.w = PyInt_AsLong(PyList_GetItem(list,10));
		}

	PyObject *getkey() {
		PyObject *list = getany();
		PyList_Append(list, PyInt_FromLong(ev.key.modifiers));
		PyList_Append(list, PyInt_FromLong(ev.key.sym));
		PyList_Append(list, PyInt_FromLong(ev.key.label));
		PyList_Append(list, PyInt_FromLong(ev.key.button));	
		return list;
		}

	void setkey(PyObject *list) {
		setany(list);
		ev.key.modifiers = PyInt_AsLong(PyList_GetItem(list,7));		
		ev.key.sym = PyInt_AsLong(PyList_GetItem(list,8));		
		ev.key.label = PyInt_AsLong(PyList_GetItem(list,9));		
		ev.key.button = PyInt_AsLong(PyList_GetItem(list,10));		
		}

	PyObject *getpmove(){
		PyObject *list = getany();
		PyList_Append(list, PyInt_FromLong(ev.pmove.x));	
		PyList_Append(list, PyInt_FromLong(ev.pmove.y));	
		PyList_Append(list, PyInt_FromLong(ev.pmove.z));	
		PyList_Append(list, PyInt_FromLong(ev.pmove.wheel));	
		return list;				
		}
	void setpmove(PyObject *list){
		setany(list);
		ev.pmove.x = PyInt_AsLong(PyList_GetItem(list,7));				
		ev.pmove.y = PyInt_AsLong(PyList_GetItem(list,8));				
		ev.pmove.z = PyInt_AsLong(PyList_GetItem(list,9));				
		ev.pmove.wheel = PyInt_AsLong(PyList_GetItem(list,10));				
		}

	PyObject *getpbutton(){
		PyObject *list = getany();
		PyList_Append(list, PyInt_FromLong(ev.pbutton.button));			
		return list;
		}
	void setpbutton(PyObject *list){
		setany(list);
		ev.pbutton.button= PyInt_AsLong(PyList_GetItem(list,7));
		}	
};
*/


class gii
{
public:
	gii_input_t inp;

	gii() { inp = NULL; }

	void Open(char *input = NULL) {
		if (input)
			inp = giiOpen(input,NULL);
		else
			inp = giiOpen(NULL,NULL);
		if ( !inp ) {
			E("Unable to initialze input !!");
			}		
		}

	~gii() {
		if ( inp )
			check(giiClose(inp));
		}
};


class visual
{
public:
	ggi_visual_t vis;


	visual() { vis = NULL; }


	ggi_pixelformat GetPixelFormat() {
		return *(ggiGetPixelFormat(vis));		
		}


	int Flush() {
		return check(ggiFlush(vis));
		}
	int FlushRegion(int x, int y, int w, int h) {
		return check(ggiFlushRegion(vis,x,y,w,h));
		}

	
	/* Graphics context
	 */
	int SetGCForeground(long  color)
	{
		return check(ggiSetGCForeground(vis,(ggi_pixel)color));
	};
	long GetGCForeground() 
	{
		ggi_pixel color;
		check(ggiGetGCForeground(vis,&color));
		return color;
	}
	int SetGCBackground(long  color)
	{
		return check(ggiSetGCBackground(vis,(ggi_pixel)color));
	};
	long GetGCBackground() 
	{
		ggi_pixel color;
		check(ggiGetGCBackground(vis,&color));
		return color;
	}
	
	int SetGCClipping(int  left,int  top,int  right,int  bottom)
	{
		return check(ggiSetGCClipping(vis,left,top,right,bottom));
	};

	PyObject *GetGCClipping()
	{
		PyObject *list = PyList_New(4);
		int left, top, right, bottom;		

		check(ggiGetGCClipping(vis,&left,&top,&right,&bottom));

		PyList_SetItem(list, 0, PyInt_FromLong(left));		
		PyList_SetItem(list, 1, PyInt_FromLong(top));		
		PyList_SetItem(list, 2, PyInt_FromLong(right));		
		PyList_SetItem(list, 3, PyInt_FromLong(bottom));		

		return list;
	};

	long MapColor(ggi_color *c)
	{
		return ggiMapColor(vis,c);
	}

	int UnmapPixel(long pix, ggi_color *c)
	{
		return check(
			ggiUnmapPixel(vis,(ggi_pixel) pix,c)
			);
	}


	// not sure on how much memory to allocate so pack and unpack are
	// not supported

	int SetPalette(int s, PyObject *py_cmap) 
	{
		ggi_color *cmap;
		int i, size = PyList_Size(py_cmap);

		cmap = new ggi_color[ size ];
		// copy into cmap
		for (i = 0; i < size; i++) {
			PyObject *pcolor = PyList_GetItem(py_cmap,i);
			PyObject *pr = PyObject_CallMethod(pcolor,"getr",NULL);
			PyObject *pg = PyObject_CallMethod(pcolor,"getg",NULL);
			PyObject *pb = PyObject_CallMethod(pcolor,"getb",NULL);
			PyObject *pa = PyObject_CallMethod(pcolor,"geta",NULL);

			cmap[i].r = PyInt_AsLong(pr);
			cmap[i].g = PyInt_AsLong(pg);
			cmap[i].b = PyInt_AsLong(pb);
			cmap[i].a = PyInt_AsLong(pa);
			}
		
		int err = check(
			ggiSetPalette(vis,s,size,cmap)
			);
		
		delete[] cmap;
		return err;
	}

//int ggiGetPalette(ggi_visual_t vis,int s,int len,ggi_color *cmap);

	PyObject *GetPalette(int s, int len) {
		ggi_color *cmap = new ggi_color[len];
		PyObject *list = PyList_New(len);
		
		check(ggiGetPalette(vis,s,len,cmap));

		for(int i = 0; i < len; i++) {
			PyObject *t = PyList_New(4);
			PyList_SetItem(t,0,PyInt_FromLong(cmap[i].r));
			PyList_SetItem(t,0,PyInt_FromLong(cmap[i].g));
			PyList_SetItem(t,0,PyInt_FromLong(cmap[i].b));
			PyList_SetItem(t,0,PyInt_FromLong(cmap[i].a));
			PyList_SetItem(list,i,t);
			}	
		delete[] cmap;

		return list;
	}


	int SetColorfulPalette() {
		return check(
			ggiSetColorfulPalette(vis)
			);
	}
//int ggiSetColorfulPalette(ggi_visual_t vis);


/* Gamma map manipulation
 */
	// return a tuple (r,g,b)
	PyObject *GetGamma() {
		ggi_float r,g,b;
		PyObject *result = PyList_New(3);
		check( ggiGetGamma(vis,&r,&g,&b) );
		PyList_SetItem(result,0,PyFloat_FromDouble(r) );
		PyList_SetItem(result,0,PyFloat_FromDouble(g) );
		PyList_SetItem(result,0,PyFloat_FromDouble(b) );
		return result;
	} 

//int ggiGetGamma(ggi_visual_t vis,ggi_float *r,ggi_float *g,ggi_float *b);

	int SetGamma(double r, double g,double b) {
		return check(
			ggiSetGamma(vis,r,g,b)
			);		
	}

	int SetGammaMap(int s, PyObject *py_gammamap) 
	{
		ggi_color *gammamap;
		int i, size = PyList_Size(py_gammamap);

		gammamap = new ggi_color[ size ];
		// copy into cmap
		for (i = 0; i < size; i++) {
			PyObject *pcolor = PyList_GetItem(py_gammamap,i);
			PyObject *pr = PyObject_CallMethod(pcolor,"getr",NULL);
			PyObject *pg = PyObject_CallMethod(pcolor,"getg",NULL);
			PyObject *pb = PyObject_CallMethod(pcolor,"getb",NULL);
			PyObject *pa = PyObject_CallMethod(pcolor,"geta",NULL);

			gammamap[i].r = PyInt_AsLong(pr);
			gammamap[i].g = PyInt_AsLong(pg);
			gammamap[i].b = PyInt_AsLong(pb);
			gammamap[i].a = PyInt_AsLong(pa);
			}
		
		int err = check(
			ggiSetGammaMap(vis,s,size,gammamap)
			);
		
		delete[] gammamap;
		return err;
	}
	
//int ggiGetGammaMap(ggi_visual_t vis,int s,int len,ggi_color *gammamap);

	


	int SetOrigin(int x, int y) {
		return check( ggiSetOrigin(vis,x,y) );
	}	

	PyObject *GetOrigin() {
		int x, y;
		PyObject *list= PyList_New(2);
		check( ggiGetOrigin(vis,&x,&y) );
		PyList_SetItem(list,0,PyInt_FromLong(x));
		PyList_SetItem(list,1,PyInt_FromLong(y));
		return list;
	}


/* Frame handling
 */
	int SetDisplayFrame(int frameno) {
		return check( ggiSetDisplayFrame(vis,frameno) );
	}
	int SetReadFrame(int frameno) {
		return check( ggiSetReadFrame(vis,frameno) );
	}
	int SetWriteFrame(int frameno){
		return check( ggiSetWriteFrame(vis,frameno) ); 
	}
	
	int GetDisplayFrame() {
		return check( ggiGetDisplayFrame(vis) );
	}
	int GetReadFrame() {
		return check( ggiGetReadFrame(vis) );
	}
	int GetWriteFrame() {
		return check( ggiGetWriteFrame(vis) );
	}

/* Generic drawing routines 
 */
	int Fillscreen() {
	return check( ggiFillscreen(vis) );
	}
	int DrawPixel(int x, int y) {
	return check( ggiDrawPixel(vis,x,y) );
	}
	int PutPixel(int x,int y, long pixel){
	return check( ggiPutPixel(vis,x,y,(ggi_pixel) pixel));
	}
	PyObject *GetPixel(int x, int y){
	long pixel;
	check(ggiGetPixel(vis,x,y,(ggi_pixel *) &pixel));
	return PyInt_FromLong(pixel);
	}
		
	int DrawLine(int x,int y,int xe,int ye)
	{
	return check(ggiDrawLine(vis,x,y,xe,ye) );
	}	
	int DrawHLine(int x,int y,int w)
	{
	return check(ggiDrawHLine(vis,x,y,w));
	}
	int DrawVLine(int x, int y, int h){
	return check(ggiDrawVLine(vis,x,y,h));
	}
	int DrawBox(int x,int y,int w,int h){
	return check(ggiDrawBox(vis,x,y,w,h));
	}	

	int CopyBox(int x,int y,int w,int h,int nx,int ny)
	{
	return check(ggiCopyBox(vis,x,y,w,h,nx,ny));	
	}

	int CrossBlit(int sx,int sy,int w,int h,
		 visual *pdst,int dx,int dy)
	{
	return check(
		ggiCrossBlit(vis,sx,sy,w,h,pdst->vis,dx,dy)
		);	
	}


//TODO
/*
// Requires that python handle memory directy
*/

	int PutBox(int x,int y,int w,int h,char *buf)
	{
		return check(ggiPutBox(vis,x,y,w,h,buf));
	};

	PyObject *GetBox(int x,int y,int w,int h,char *buf, long buflen)
	{
		check(ggiPutBox(vis,x,y,w,h,buf));
		return PyString_FromStringAndSize(buf,buflen);
	};

	int PutHLine(int x, int y, int w, char *buf) {
		return check(ggiPutHLine(vis,x,y,w,(void *)buf));
	}

	PyObject *GetHLine(int x, int y, int w, char *buf, long buflen) {
		check(ggiGetHLine(vis,x,y,w,(void *)buf));
		return PyString_FromStringAndSize(buf,buflen);
	}


	int PutVLine(int x, int y, int w, char *buf) {
		return check(ggiPutVLine(vis,x,y,w,(void *)buf));
	}

	PyObject *GetVLine(int x, int y, int w, char *buf, long buflen) {
		check(ggiGetVLine(vis,x,y,w,(void *)buf));
		return PyString_FromStringAndSize(buf,buflen);
	}


/* Text drawing routines 
*/
	int Putc(int x,int y,char c)
	{
	return check(
		ggiPutc(vis,x,y,c)
		);
	}
	int Puts(int x,int y,char *c)
	{
	return check(
		ggiPuts(vis,x,y,c)
		);
	}

	PyObject *GetCharSize() {
	PyObject *list = PyList_New(2);
	int width, height;
	check(ggiGetCharSize(vis,&width,&height));
	PyList_SetItem(list,0,PyInt_FromLong(width));
	PyList_SetItem(list,1,PyInt_FromLong(height));
	return list;
	}

	gii JoinInputs(gii *pgii)  {
		gii_input_t t;
		gii g;
		t = ggiJoinInputs(vis,pgii->inp);
		g.inp = t;
		return g;				
		}

	int EventSend(ggi_event *e) {
		int check(ggiEventSend(vis, e));
	}

	ggi_event EventRead(long mask)
	{
	ggi_event e;
	check(ggiEventRead(vis, &e,(gii_event_mask) mask));
	return e;
	}

	int AddEventMask(long evm)
	{
	return check(ggiAddEventMask(vis,(gii_event_mask)evm));
	}

	int RemoveEventMask(long evm)
	{
	return check(ggiRemoveEventMask(vis,(gii_event_mask)evm));
	}
	

	long SetEventMask(long evm)
	{
	return check(ggiSetEventMask(vis, (gii_event_mask) evm));
	}

	long GetEventMask() 
	{
	return 	ggiGetEventMask(vis);
	}
	
	int EventsQueued(long mask)
	{
	return check( ggiEventsQueued(vis, (gii_event_mask) mask) );
	}
	

//	long EventPoll(long mask, long seconds, long microseconds){
	long EventPoll(long mask, struct timeval *t){
//	struct timeval t;

//	t.tv_usec = microseconds;
//	t.tv_sec = seconds;

	return ggiEventPoll(vis, (gii_event_mask) mask, t);
	} 
	

	int Kbhit() { return ggiKbhit(vis); }
	int Getc() { return ggiGetc(vis); }



	// set/get mode object
	int SetMode(ggi_mode *tm) {
		return check(
			ggiSetMode(vis, tm) );
		}

	ggi_mode GetMode() {
		ggi_mode tm;
	
		check(ggiGetMode(vis,&tm));

		return tm;
		}

	
	int CheckMode(ggi_mode *tm)
	{
		return ggiCheckMode(vis, tm);	
	};

	int SetTextMode(int cols,int rows,
			   int vcols,int vrows,int fontx,int fonty,
			   long type)
	{
		return check(
			ggiSetTextMode(vis,cols,rows,vcols,vrows,fontx,
				fonty, (ggi_graphtype) type) );
	};

	ggi_mode CheckTextMode(int cols,int rows,
			     int vcols,int vrows, int fontx,int fonty,
			      long type)
	{
		ggi_mode tm;
		int err;

		err = ggiCheckTextMode(vis,cols,rows,
			     vcols,vrows,fontx,fonty,
				(ggi_graphtype) type, &tm);
		check(err);

		return tm;
	}

	int SetGraphMode(int x, int y,
			    int xv,int yv,long type)
	{
	return check( ggiSetGraphMode(vis, x, y,
			    xv, yv,(ggi_graphtype) type));
	};

	ggi_mode CheckGraphMode(int x, int y,
			      int xv, int yv, long type
			      )
	{
		ggi_mode tm;
		int err;

		err = ggiCheckGraphMode(vis,x,y,xv,yv,
				(ggi_graphtype) type, &tm);
		check(err);

		return tm;		
	}


	int SetSimpleMode(int xsize, int ysize, int frames,
			     long type)
	{
		return check(
			ggiSetSimpleMode(vis, xsize,ysize,frames,
			     (ggi_graphtype) type) );
	}


	ggi_mode CheckSimpleMode(int xsize, int ysize, int frames,
			       long type)
	{
		ggi_mode tm;
		int err;

		err = ggiCheckSimpleMode(vis,xsize,ysize,frames,
				(ggi_graphtype) type, &tm);
		check(err);

		return tm;			
	}

	// flag management
	int SetFlags(long flags) {
		return check(ggiSetFlags(vis,(ggi_flags) flags));
		}	

	long GetFlags() {
		return ggiGetFlags(vis);
		}
	
	int AddFlags(long flags) {
		return ggiSetFlags(vis,flags);
		}

	int RemoveFlags(long flags) {
		return ggiRemoveFlags(vis,flags);
		}
	

};



class ggi
{
public:
	ggi() {
		ggiInit();
	}

	~ggi() {
		ggiExit();
	}

	void Panic(char *msg) {
		ggiPanic(msg);
	}

	visual Open(char *dname = NULL) {
		visual v;
		if (dname == NULL) {
			v.vis = ggiOpen(NULL,NULL);
			if (v.vis == NULL) {
				E("ggiOpen returned null!");
				}
			} else {
			v.vis = ggiOpen(dname);
			if (v.vis == NULL) {
				char msg[81];
				sprintf(msg,"ggiOpen(%s) returned null",dname);
				E(msg);
				}
			}
		return v;
	}// end Open

	int Close(visual *v) {
		int err = ggiClose(v->vis);
		return check(err);		
	} // end Close 

	
};


















%}
